sTitle = 'Theme Assembler'
reaper.ClearConsole()

OS = reaper.GetOS()
script_path = ({reaper.get_action_context()})[2]:match('^.*[/\\]'):sub(1,-2)
themes_path = reaper.GetResourcePath() .. "/ColorThemes"

if not reaper.file_exists(script_path..'/theme_assembler/initial/Assembled_Theme.ReaperTheme') then
  reaper.MB("theme_assembler path does not exist alongside script file in "..script_path,
            "Theme Assembler Installation Error",0);
else


gfx.init(sTitle,
tonumber(reaper.GetExtState(sTitle,"wndw")) or 800,
tonumber(reaper.GetExtState(sTitle,"wndh")) or 600,
tonumber(reaper.GetExtState(sTitle,"dock")) or 0,
tonumber(reaper.GetExtState(sTitle,"wndx")) or 100,
tonumber(reaper.GetExtState(sTitle,"wndy")) or 50)

gfx.ext_retina = 1







  ---------- COLOURS -----------
  
function setCol(col)
  if col[1] and col[2] and col[3] then
    local r = col[1] / 255
    local g = col[2] / 255
    local b = col[3] / 255
    local a = 1
    if col[4] ~= nil then a = col[4] / 255 end
    gfx.set(r,g,b,a)
    gfx.a2=gfx.a
  end
end

function getProjectCustCols()
  projectCustCols = {}
  for i=0, reaper.CountTracks(0)-1 do
    local c = reaper.GetTrackColor(reaper.GetTrack(0, i))
    if c ~= 0 then
      local r,g,b = reaper.ColorFromNative(c)
      table.insert(projectCustCols,{r,g,b})
    end
  end
end

representCustCol = {70,90,90}

  ---------- TEXT -----------

textPadding = 6

if OS:find("Win") ~= nil then

  gfx.setfont(1, "Calibri", 13)
  gfx.setfont(2, "Calibri", 15)
  gfx.setfont(3, "Calibri", 22)
  gfx.setfont(4, "Calibri Bold", 50)
  
  gfx.setfont(11, "Calibri", 19)
  gfx.setfont(12, "Calibri", 22)
  gfx.setfont(13, "Calibri", 33)
  gfx.setfont(14, "Calibri Bold", 75)
  
  gfx.setfont(21, "Calibri", 26)
  gfx.setfont(22, "Calibri", 30)
  gfx.setfont(23, "Calibri", 44)
  gfx.setfont(24, "Calibri Bold", 100)

elseif OS == 'Other' then
  gfx.setfont(1, "Open Sans", 9)
  gfx.setfont(2, "Open Sans", 11)
  gfx.setfont(3, "Open Sans", 20)
  gfx.setfont(4, "Open Sans Bold", 40)
  
  gfx.setfont(11, "Open Sans", 13)
  gfx.setfont(12, "Open Sans", 15)
  gfx.setfont(13, "Open Sans", 30)
  gfx.setfont(14, "Open Sans Bold", 60)
  
  gfx.setfont(21, "Open Sans", 18)
  gfx.setfont(22, "Open Sans", 22)
  gfx.setfont(23, "Open Sans", 40)
  gfx.setfont(24, "Open Sans Bold", 80)

else

  gfx.setfont(1, "Helvetica", 9)
  gfx.setfont(2, "Helvetica", 11)
  gfx.setfont(3, "Helvetica", 19)
  gfx.setfont(4, "Helvetica Bold", 40)
  
  gfx.setfont(11, "Helvetica", 13)
  gfx.setfont(12, "Helvetica", 15)
  gfx.setfont(13, "Helvetica", 30)
  gfx.setfont(14, "Helvetica Bold", 75)
  
  gfx.setfont(21, "Helvetica", 18)
  gfx.setfont(22, "Helvetica", 22)
  gfx.setfont(23, "Helvetica", 38)
  gfx.setfont(24, "Helvetica Bold", 92)
end

function text(str,x,y,w,h,align,col,style,lineSpacing,vCenter,wrap)
  local lineSpace = lineSpacing or (11*scaleMult)
  setCol(col or {255,255,255})
  gfx.setfont(style or 1)

  local lines = nil
  if wrap == true then lines = textWrap(str,w)
  else
    lines = {}
    for s in string.gmatch(str, "([^#]+)") do
      table.insert(lines, s)
    end
  end
  if vCenter ~= false and #lines > 1 then y = y - lineSpace/2 end
  for k,v in ipairs(lines) do
    gfx.x, gfx.y = x,y
    gfx.drawstr(v,align or 0,x+(w or 0),y+(h or 0))
    y = y + lineSpace
  end
end

function textWrap(str,w) -- returns array of lines
  local lines,curlen,curline,last_sspace = {}, 0, "", false
  -- enumerate words
  for s in str:gmatch("([^%s-/]*[-/]* ?)") do
    local sspace = false -- set if space was the delimiter
    if s:match(' $') then
      sspace = true
      s = s:sub(1,-2)
    end
    local measure_s = s
    if curlen ~= 0 and last_sspace == true then
      measure_s = " " .. measure_s
    end
    last_sspace = sspace

    local length = gfx.measurestr(measure_s)
    if length > w then
      if curline ~= "" then
        table.insert(lines,curline)
        curline = ""
      end
      curlen = 0
      while length > w do -- split up a long word, decimating measure_s as we go
        local wlen = string.len(measure_s) - 1
        while wlen > 0 do
          local sstr = string.format("%s%s",measure_s:sub(1,wlen), wlen>1 and "-" or "")
          local slen = gfx.measurestr(sstr)
          if slen <= w or wlen == 1 then
            table.insert(lines,sstr)
            measure_s = measure_s:sub(wlen+1)
            length = gfx.measurestr(measure_s)
            break
          end
          wlen = wlen - 1
        end
      end
    end
    if measure_s ~= "" then
      if curlen == 0 or curlen + length <= w then
        curline = curline .. measure_s
        curlen = curlen + length
      else -- word would not fit, add without leading space and remeasure
        table.insert(lines,curline)
        curline = s
        curlen = gfx.measurestr(s)
      end
    end
  end
  if curline ~= "" then
    table.insert(lines,curline)
  end
  return lines
end



  --------- IMAGES ----------
  
imgBufferOffset = 500  

function loadImage(idx, name, location, noIScales)
  local i = idx
  if i then
    if location then
      str = script_path..'/theme_assembler/'..location.."/"..name..".png"
      if gfx.loadimg(i, str) == -1 then reaper.ShowConsoleMsg("script image "..name.." not found\n") end
    else
      local scaleFolder = ''
      if noIScales ~= true and scaleMult == 1.5 then scaleFolder = '/150' end
      if noIScales ~= true and scaleMult == 2 then scaleFolder = '/200' end
      local str = themes_path..'/'..themeFolder..scaleFolder.."/"..name..".png"
      if gfx.loadimg(i, str) == -1 then reaper.ShowConsoleMsg("theme image "..name.." not found\n") end
    end
  end
end


function getImage(img, location, noIScales)
  local buf = nil
  local i = imgBufferOffset
  while buf == nil do
    local h,w = gfx.getimgdim(i)
    if h==0 then buf=i end
    i = i+1
  end
  --reaper.ShowConsoleMsg('image to '..buf..'\n')
  loadImage(buf, img, location, noIScales) 
  return buf
end

function copyImg(el) 
  for i,v in ipairs(el.states) do 
  local scales = {{'',''},{'_150','/150'},{'_200','/200'}}
    if el.iType == 'meterV' or el.iType == 'meterH' then scales = {{'',''}} end
    for p,q in ipairs(scales)do
      --reaper.ShowConsoleMsg('src is '..path.."Scripts\\theme_assembler\\"..v[1]..'\\'..el.img..q[1]..'.png\n')
      --reaper.ShowConsoleMsg('dest is '..path.."ColorThemes\\"..themeFolder..q[2]..'\\'..v[1]..'.png\n')
      src = script_path.."/theme_assembler/"..v[1]..'/'..el.img..q[1]..'.png'
      dest = themes_path.."/"..themeFolder..q[2]..'/'..v[1]..'.png'
      osCopy(src, dest)
    end
  end
  refreshTheme()
  reloadImgs()
end

function reloadImgs()
  for b in ipairs(els) do -- iterate blocks 
    for z in ipairs(els[b]) do -- iterate z
      if els[z] ~= nil then
        for j,k in pairs(els[b][z]) do
          k:reloadImg()
        end
      end
    end
  end
  doArrange = true
end

  --------- OBJECTS ----------

els = {}
function AddEl(o)
  if o.x == nil and o.y == nil and o.updateOn == nil and o.action == nil then --just a proto
  else
    if o.parent then  adoptChild(o.parent, o) end
    if o.block == nil then if o.parent and o.parent.block then o.block = o.parent.block  end end-- no block specified, inherit from parent
    if o.z == nil then if o.parent and o.parent.z then o.z = o.parent.z  end end -- no z specified, inherit from parent
    if els[o.block] == nil then els[o.block] = {} end
    if els[o.block][o.z] == nil then els[o.block][o.z] = {o}
    else els[o.block][o.z][#els[o.block][o.z]+1] = o
    end
  end
end

El = {}
function El:new(o)
  local o = o or {}
  if o.interactive == nil then o.interactive = true end
  self.__index = self
  AddEl(o)
  setmetatable(o, self)
  return o
end

Block = {}
function Block:new(o)
  local o = o or {}
  if els == nil then els = {} end
  els[#els+1] = o
  self.__index = self
  setmetatable(o, self)
  return o
end

Button = El:new{} 
function Button:new(o)
  self.__index = self
  AddEl(o)
  setmetatable(o, self)
  return o
end

funcText = El:new{} 
function funcText:new(o)
  self.__index = self
  AddEl(o)
  o.onArrange = function() self.w = fitText(o)+16 end
  setmetatable(o, self)
  return o
end

TargetButton = El:new{} 
function TargetButton:new(o)
  self.__index = self
  AddEl(o)
  o.z = o.z or 2
  o.iType = o.iType or 3
  if o.iType == 'stack' then o.iFrameH, o.iFrame = o.iFrameH or 20, o.iFrame or 1 end
  if o.iType == ('meterV' or 'meterH') then o.mValue = 0 end
  o.img = o.title or nil
  if o.onClick == nil then
    o.onClick = function() setTarget(o) end
  end
  if o.iType == 'themeFile' then
      o.children = {
      El:new({parent=o, x=0, y=0, w=110, h=50, onClick=o.onClick}),
      El:new({parent=o.children[1], x=1, y=1, w=48, h=48, img="reaper_theme_icon", iLocation='scriptImages', iType='scriptImage', onClick=o.onClick}),
      El:new({parent=o.children[1], x=44, y=14, w=60, h=38, onClick=o.onClick, text={str='Theme Colors & Fonts File', style=1,align=3,wrap=true, col={200,200,200}}})
      }
  end
  if o.iType == 'idxbgColorHack' then
    o.w, o.h = 64,50
    o.col = representCustCol
  end
  setmetatable(o, self)
  return o
end

ChoiceBox = El:new{} 
function ChoiceBox:new(ch) 
  self.__index = self
  local title = ch.img or ch.title
  o = El:new({parent=ch.parent, flow=ch.flow or false,  resizeParent='v', text={str=string.gsub(title, "_", " "),style=1,align=9,col={255,255,255}},
        x=0, y=0, border=10, w=ch.w or 80, h=ch.h or 50, col={70,70,70}, mouseOverCol={80,80,80},
        onClick = ch.onClick or function() copyImg(ch) end})
  AddEl(o)
  o.children = {}
  if ch.iType == 'stack' then
    ch.iFrameH, ch.iFrame = ch.iFrameH or ch.w, ch.iFrame or 1
  end
  ch.iType = ch.iType or 3
  o.children.img = El:new({parent=o, x=0, y=-7, iFrameH = 20, iFrame = ch.iFrame,
                elAlign={x={'parent','centre'},y={'parent','centre'}}, img=ch.img, iType=ch.iType, noIScales=ch.noIScales, onFps=ch.onFps or nil, 
                imgPink=ch.imgPink, iLocation=ch.iLocation, states=ch.states, meterColumn=ch.meterColumn, onClick = o.onClick})
  if ch.iType == 'themeFile' then
    o.children.img = nil
    o.onClick = function() copyThemeFile(ch) end
    El:new({parent=o, x=0, y=0, w=100, h=36, col=ch.proxyCols[1], onClick=o.onClick})
    El:new({parent=o, x=4, y=4, w=24, h=28, col=ch.proxyCols[2], onClick=o.onClick})
    El:new({parent=o, x=28, y=4, w=24, h=14, col=ch.proxyCols[3], onClick=o.onClick})
    El:new({parent=o, x=28, y=18, w=24, h=14, col=ch.proxyCols[4], onClick=o.onClick})
  end
  if ch.iType == 'idxbgColorHack' then
    o.col = representCustCol
  end
  setmetatable(o, self)
  return o
end

OrderField = El:new{} 
function OrderField:new(o)
  self.__index = self
  o.parent = orderThing
  o.x, o.y, o.w, o.h = 8, 8, o.w or 120, 54
  o.col={60,60,60}
  --o.col={255,0,0}
  AddEl(o)
  o.children = {} 
  if o.lButton then 
    local l = Button:new({parent=o, x=-8, elAlign={x={'parent','centre'}}, y=30, w=16, h=24, img='arrow_left', iType=3, iLocation='scriptImages',
      onClick=function() 
        swapFields(o.index-1,o.index)
      end})
    table.insert(o.children, l)
  end
  if o.rButton then 
    local r = Button:new({parent=o, x=8, elAlign={x={'parent','centre'}}, y=30, w=16, h=24, img='arrow_right', iType=3, iLocation='scriptImages', 
      onClick=function() 
        swapFields(o.index,o.index+1)
      end})
    table.insert(o.children, r)
  end
  
  setmetatable(o, self)
  return o
end




  --------- FUNCS ----------
  
  
function El:purge()
  --reaper.ShowConsoleMsg('purging\n')
  for b in ipairs(els) do -- iterate blocks
    for z in ipairs(els[b]) do
      if els[b][z] ~= nil and #els[b][z] ~= 0 then
        for j,k in pairs(els[b][z]) do
          if k == self then
            if self.children ~= nil then 
              for l,m in pairs(self.children) do
                m:purge() 
              end 
            end
            if self:addToDirtyZone(b,z) == true then
              if self.imgIdx then
                gfx.setimgdim(self.imgIdx,0,0)
              end
              table.remove(els[b][z],j)
              doDraw = true
            end
          end
        end
      end
    end
  end -- end iterating blocks
end

function colCycle(self) -- for debugging / inducing headaches
  if colDebug ~= true then self.col = {0,255,0,150}
  else self.col = {math.random(255),math.random(255),math.random(255),255}
    self:addToDirtyZone()
  end
end  

function osCopy(src, dest)

  if src:sub(-4) == ".png" or  src:sub(-12) == ".ReaperTheme" then
    local srcf = io.open(src,"rb")
    if srcf then
      local dstf = io.open(dest,"wb")
      if dstf then
        while true do
          local b = srcf:read(32768)
          if not b then break end
          dstf:write(b)
        end
      end
      srcf:close()
      if dstf then dstf:close() return true end
    end
  end
  -- paths should have forward slashes for everything here
  if OS:find("Win") ~= nil then
    src = src:gsub("/","\\") -- xcopy wants backslashes
    dest = dest:gsub("/","\\")
    --reaper.ShowConsoleMsg('xcopy /e/f/y "'..src..'" "'..dest..'" \n')
    os.execute('xcopy /e/f/i/y "'..src..'" "'..dest..'"')
  else
    assert(src:match("\\")==nil, "src for filecopy has backslash")
    assert(dest:match("\\")==nil, "dest for filecopy has backslash")
    os.execute('cp -fR "'..src..'" "'..dest .. '"')
  end
  return true
end

function tableOfDir(folder)
  local filelist=io.popen('dir "' .. folder .. '" /b')
  local t ={}
  for filename in filelist:lines() do
    --reaper.ShowConsoleMsg(filename ..'\n')
    table.insert(t, filename)
  end
  return t
end
--steve = tableOfDir(script_path..'/theme_assembler')

function getRtconfigVar(var)
  local str = themes_path..'/'..themeFolder..'/'..'rtconfig.txt'
  for line in io.lines(str) do
    if string.find(line, 'set '..var) then
      return string.match(line, "%d+")
    end
  end
end

function setRtconfigVar(var, val)
  local str = themes_path..'/'..themeFolder..'/'..'rtconfig.txt'
  local rtConfigTable = {}
  
  for line in io.lines(str) do
    if string.find(line, 'set '..var) then
      table.insert(rtConfigTable, 'set '..var..' '..val)
    else
      table.insert(rtConfigTable,line)
    end
  end
  
  local file = io.open(str, "w")
  io.output(file)
  for i,v in ipairs(rtConfigTable) do
    io.write(v..'\n')
  end
  io.close(file)
  refreshTheme()
  
end

function swapFields(f1,f2)
 local str = themes_path..'/'..themeFolder..'/'..'rtconfig.txt'
  local rtConfigTable = {}
  for line in io.lines(str) do table.insert(rtConfigTable,line) end
  
  local file = io.open(str, "w")
  io.output(file)
  local fl = 0
  for i,v in ipairs(rtConfigTable) do
    if string.find(v, 'then tcp') then
      fl = fl + 1
      if fl >= 11 then fl = 1 end
      if fl == f1 then io.write(rtConfigTable[i+1]..'\n')
        elseif fl == f2 then io.write(rtConfigTable[i-1]..'\n')
        else io.write(v..'\n')
      end 
    else
      io.write(v..'\n') -- unchanged line
    end
  end  
  io.close(file)
  refreshTheme()
  doArrange = true
end


function math.Clamp(val, min, max)
  return math.min(math.max(val, min), max)
end


function copyThemeFile(el) 
  src = script_path.."/theme_assembler/ReaperTheme/"..el.title..'.ReaperTheme'
  --reaper.ShowConsoleMsg('src is '..src..'\n')
  dest = themes_path..'/Assembled_Theme.ReaperTheme'
  --reaper.ShowConsoleMsg('dest is '..dest..'\n')
  osCopy(src, dest)
  refreshTheme()
end

function themeCheck()
  if themeTitle ~= oldthemeTitle or themeTitle == nil then
    last_theme_filename = reaper.GetLastColorThemeFile()
    last_theme_filename_check = reaper.time_precise()
    themeTitle = string.match(string.match(last_theme_filename, '[^\\/]*$'),'(.*)%..*$')
    switchTheme(themeTitle)
    oldthemeTitle = themeTitle
  else 
    local now = reaper.time_precise()
    if now > last_theme_filename_check+1 and suspendSwitchTheme ~= true  then -- once per second see if the theme filename changed
      last_theme_filename_check = now
      local tfn = reaper.GetLastColorThemeFile()
      if tfn ~= last_theme_filename then
        themeTitle = string.match(string.match(tfn, '[^\\/]*$'),'(.*)%..*$')
        switchTheme(themeTitle)
        last_theme_filename = tfn
      end
    end
  end
end

function switchTheme(thisTheme)
  if thisTheme ~= 'Assembled_Theme' and script_path and themes_path then
    if(reaper.file_exists(themes_path..'/Assembled_Theme.ReaperTheme')==true) then
      local v = reaper.ShowMessageBox('switch to Assembled Theme?', 'Wrong theme in use', 1)
      if v == 1 then reaper.OpenColorThemeFile(themes_path..'/Assembled_Theme.ReaperTheme') end
      if v == 2 then gfx.quit() end
    else
      local v = reaper.ShowMessageBox('Create theme?', 'Assembled Theme not found', 1)
      if v == 1 then 
        if OS:find("Win") ~= nil then
          if osCopy(script_path..'/theme_assembler/initial/', themes_path..'/') == true 
             then reaper.OpenColorThemeFile(themes_path..'/Assembled_Theme.ReaperTheme') end
        else
          if osCopy(script_path..'/theme_assembler/initial/Assembled_Theme.ReaperTheme', themes_path..'/Assembled_Theme.ReaperTheme') == true and
             osCopy(script_path..'/theme_assembler/initial/Assembled_Theme_Resources', themes_path..'/Assembled_Theme_Resources') == true
            then reaper.OpenColorThemeFile(themes_path..'/Assembled_Theme.ReaperTheme') end
        end
      end
      if v == 2 then gfx.quit() end
    end
  end
end

function setTarget(n) 
  target = n.title
  populateTarget(n)
end

function adoptChild(parent, child)
  if parent.children then parent.children[#parent.children + 1] = child
  else parent.children = {child}
  end
end

function populateTarget(t)

  if chooserBox then chooserBox:purge() end -- kill a previous one
  chooserBox = El:new({parent=thinBox, x=0, y=0, flow=mcpBox, w=-10, h=100, r={toEdge, thinBox, 'right'}, children={}})
  if t.states then
 
    if t.chosenState==nil then t.chosenState = {} end
    for i,v in pairs(t.states) do -- i is button states

      chooserBox.children[i] = El:new({parent=chooserBox, flow=true,  x=0, y=0, border=10, w=0, r={toEdge, chooserBox, 'right'}, h=40, col={30,30,30}})
      titleBox = El:new({parent=chooserBox.children[i], x=0, y=0, w=0, h=16, r={toEdge, chooserBox, 'right'}, col={50,50,50}})
      
      if t.iType == ('meterV' or 'meterH') then
        for q,r in ipairs({'Normal','Normal Clipped','Armed','Armed Clipped'}) do
          funcText:new({parent=titleBox, x=0, y=0, h=16, flow=true, resizeParent='v', mouseOverCol={255,255,255,50}, 
            text={str=r, style=2, align=1, col={200,200,200}},
            onClick = function() 
              t.meterColumn = q
              populateTarget(t)
            end}) -- the text that chooses from the different meter columns
         end
      
      else
        for q,r in pairs(v) do -- r[1] is image file name, r[2] is image title
          funcText:new({parent=titleBox, x=0, y=0, h=16, flow=true, resizeParent='v', mouseOverCol={255,255,255,50}, 
            text={str=string.gsub(r[2], "_", " "), style=2, align=1, col={200,200,200}},
            onClick = function() 
              t.chosenState[i] = q
              populateTarget(t)
            end}) -- the text that chooses from the different styles
        end
      end
      
      local stateHasPink = false
      if t.imgPinkStates then
        for r, state in ipairs(t.imgPinkStates) do
          if state == i then stateHasPink = true end
        end
      end
      
      local stateIType = t.iType or nil
      if t.iTypeStates then stateIType = t.iTypeStates[i] end
      local varState = t.chosenState[i] or 1 -- meters don't have chosenState
      local p = script_path.."/theme_assembler/"..v[varState][1].."/"

      for l,m in ipairs(t.buttonStyles or buttonStyles) do
        local scaleString = ''
        if scaleFactor>100 then scaleString = '_'..scaleFactor end
        if t.iType == 'themeFile' or reaper.file_exists(p..m..scaleString..'.png') then
          
          if t.iType == 'meterV' then
            ChoiceBox:new({parent=chooserBox.children[i],x=0,y=0,h=t.buttonH or nil,flow=true, img=m, noIScales=t.noIScales, iType=stateIType, onFps=t.onFps or nil, 
                            imgPink=stateHasPink, iLocation=v[1][1], meterColumn=(t.meterColumn or 1), states=v})
          elseif t.iType == 'meterH' then
            ChoiceBox:new({parent=chooserBox.children[i],x=0,y=0,w=t.buttonW or nil,flow=true, img=m, noIScales=t.noIScales, iType=stateIType, onFps=t.onFps or nil, 
                            imgPink=stateHasPink, iLocation=v[1][1], meterColumn=(t.meterColumn or 1), states=v})
          elseif t.iType == 'themeFile' then
            ChoiceBox:new({parent=chooserBox.children[i],x=0,y=0,w=100,h=50, flow=true, col={255,0,0}, title=m, iType='themeFile',
                            states=v, proxyCols=t.styleProxyCols[l]}) 
          elseif t.iType == 'idxbgColorHack' then
            ChoiceBox:new({parent=chooserBox.children[i],x=0,y=0,w=80,h=50, flow=true, col={255,0,0}, img=m, iType='idxbgColorHack',
                            iLocation=v[varState][1], states=v})
          else
            ChoiceBox:new({parent=chooserBox.children[i],x=0,y=0,h=t.buttonH or nil,flow=true, img=m, iType=stateIType, onFps=t.onFps or nil, 
                          imgPink=stateHasPink, iLocation=v[varState][1], states=v})
          end
          
        end
      end
    end
  end
  
  doArrange = true
end

function cycleBitmapStack(self)
  if self.iFrameC then
    self.iFrameDirection = self.iFrameDirection or 1
    if self.iFrame == self.iFrameC-1 and self.iFrameDirection == 1 then self.iFrameDirection = -1 end
    if self.iFrame == 0 then self.iFrameDirection = 1 end
    self.iFrame = self.iFrame + self.iFrameDirection
    self:addToDirtyZone()
  end
end

function fakeMeterVals(self)
  local newRnd, newVal = math.random(), (self.mValue or 1) * 0.85
  if newRnd > 0.8 then newVal = newRnd end
  self.mValue = newVal
  self:addToDirtyZone()
end

function refreshTheme()
  local thisTheme = reaper.GetLastColorThemeFile()
  reaper.OpenColorThemeFile(thisTheme)
end

function El:reloadImg()
  if self.img then
    local i = scaleToDrawImg(self) 
    loadImage(self.imgIdx, i, self.iLocation or nil, self.noIScales)
    self:addToDirtyZone()
  end
end

function scaleToDrawImg(self)
  local i = self.img
  local doScale = true
  if self.noIScales and self.noIScales == true then doScale = false end
    if self.iLocation and doScale == true then -- image lives in script resources
      if scaleMult == 1.5 then i = self.img..'_150' end
      if scaleMult == 2 then i = self.img..'_200' end 
    end
  return i
end

function setScale(scale)
  scaleMult = scale
  if scaleMult == 1 then textScaleOffs = 0 end
  if scaleMult == 1.5 then textScaleOffs = 10 end
  if scaleMult == 2 then textScaleOffs = 20 end
  reloadImgs()
  doArrange = true
end

function fitText(self)
  gfx.setfont(self.text.style or 1)
 return gfx.measurestr(self.text.str)
end

function getPreviousEl(self) -- returns the previous child of this element's parent
  if self.parent then
    for i=1, #self.parent.children do
      if self.parent.children[i] == self and i>1 then
        return self.parent.children[i-1] 
      end
    end
  end
end

  --------- ARRANGE ----------

function El:dirtyXywhCheck(b,z)
    if self.drawX == nil then -- then you've never been arranged
      if self:arrange(self) == true then 
        --reaper.ShowConsoleMsg('first time dirtyXywhCheck adding to dirty zone\n')
        self:addToDirtyZone(b, z, false) 
      end
    else
      self.ox,self.oy,self.ow,self.oh = self.drawX, self.drawY, self.drawW, self.drawH
      if self:arrange(self) == true then
        if self.drawX ~= self.ox or self.drawY ~= self.oy or self.drawW ~= self.ow or self.drawH ~= self.oh then 
          --reaper.ShowConsoleMsg('changed dirtyXywhCheck adding to dirty zone\n')
          self:addToDirtyZone(b, z, true)
        end
      end
    end
end

function El:addToDirtyZone(b, z, newXywh)
  b = b or self.block or 1
  z = z or self.z or 1 
  local kx,ky,kw,kh = self.drawX or self.x, self.drawY or self.y, self.drawW or self.w or 0, self.drawH or self.h or 0 
  if dirtyZones[b] == nil then dirtyZones[b] = {} end
  if dirtyZones[b][z] == nil then dirtyZones[b][z] = {x1={},y1={},x2={},y2={}} end
  if kw ~= nil then
    dirtyZones[b][z].x1[#dirtyZones[b][z].x1+1] = kx
    dirtyZones[b][z].y1[#dirtyZones[b][z].y1+1] = ky
    dirtyZones[b][z].x2[#dirtyZones[b][z].x2+1] = kx + kw
    dirtyZones[b][z].y2[#dirtyZones[b][z].y2+1] = ky + kh
    --reaper.ShowConsoleMsg('b='..b..',z='..z..' x y w h = '..kx..' '..ky..' '..kw..' '..kh..'\n')
  end
  if newXywh == true then -- element has moved, so also dirtyZone its old location
    dirtyZones[b][z].x1[#dirtyZones[b][z].x1+1] = self.ox
    dirtyZones[b][z].y1[#dirtyZones[b][z].y1+1] = self.oy
    dirtyZones[b][z].x2[#dirtyZones[b][z].x2+1] = self.ox + self.ow
    dirtyZones[b][z].y2[#dirtyZones[b][z].y2+1] = self.oy + self.oh
  end
  doDraw = true
  return true
end

function El:drawToBuffer(buffer,dx,dy,dw,dh)
  local kx,ky,kw,kh = self.drawX or self.x, self.drawY or self.y, self.drawW or self.w or 0, self.drawH or self.h or 0 
  if hasOverlap(kx,ky,kw,kh,dx,dy,dw,dh) == true then
    self:draw(buffer,dx,dy) --draw the element to the buffer
  end
end

function hasOverlap(x1,y1,w1,h1,x2,y2,w2,h2)
  if x1+w1 > x2 and x1 < x2 + w2 and y1+h1 > y2 and y1 < y2 + h2 then return true end
end

function resizeParent(self)
  if self.resizeParent and self.resizeParent=='v' and (self.drawY + self.drawH) > (self.parent.drawY + self.parent.drawH) then
    self.parent.ox, self.parent.oy, self.parent.ow, self.parent.oh = self.parent.drawX, self.parent.drawY, self.parent.drawW, self.parent.drawH
    self.parent.drawH = self.drawY + self.drawH - self.parent.drawY + (self.border or 0)
    self.parent:addToDirtyZone(self.parent.block or 1, self.parent.z or 1, true)
    if self.parent.parent then resizeParent(self.parent) end
  end
end

function toEdge(self,edge) -- sets an edge to another element's edge. Called by el:arrange()
 if edge == 'left' then -- my left edge
    if self.l[3] == 'left' then reaper.ShowConsoleMsg('left toEdge left not done yet\n') end
    if self.l[3] == 'right' then return self.l[2].drawX + self.l[2].drawW + self.x end
  end
  if edge == 'top' then -- my top edge
    if self.t[3] == 'top' then return self.t[2].drawY + self.y end
    if self.t[3] == 'bottom' then return self.t[2].drawY + self.t[2].drawH + self.y end
  end
  if edge == 'right' then -- my right edge
    if self.r[3] == 'left' then reaper.ShowConsoleMsg('right toEdge left not done yet\n') end
    if self.r[3] == 'right' then return self.r[2].drawX + self.r[2].drawW - (self.drawX or self.x) + self.w end
  end
  if edge == 'bottom' then -- my bottom edge
    if self.b[3] == 'top' then reaper.ShowConsoleMsg('bottom toEdge top not done yet\n') end
    if self.b[3] == 'bottom' then return self.b[2].drawY + self.b[2].drawH - self.drawY + self.h end
  end
end

function El:arrange()
  local px, py, pw, ph, pCOffx, pCOffy = 0, 0, 0, 0, 0, 0 -- pCOff are the parent's clipping offsets
  if self.parent ~= nil then 
    px, py, pw, ph = self.parent.drawX or 0, self.parent.drawY or 0, self.parent.drawW or 0, self.parent.drawH or 0 
    --if self.parent.clipOffset then pCOffx, pCOffy = self.parent.clipOffset[1], self.parent.clipOffset[2] end
  end
  
  if self.fieldParent ~= nil then
    for i,v in ipairs(orderFields) do
      if v.field == self.fieldParent then
        px, py, pw, ph = v.drawX or v.x or 0, v.drawY or v.y or 0, v.drawW or 0, v.drawH or 0
        v.w = self.fieldW or 120
      end
    end
  end
  
  if self.onArrange then self.onArrange(self) end
  
  if self.img ~= nil and (self.w == nil or self.h == nil) then
    if self.iType == 'meterV' then self.w, self.h = 12,100 
    elseif self.iType == 'meterH' then self.w, self.h = 100,12
    elseif self.iType == 'idxbgColorHack' then self.w, self.h = 74,30
    else self.measureImg = true end -- w or h not provided, use image size
  end
  
  self.drawX, self.drawY = px+((self.x or 0)+(self.border or 0)-pCOffx)*scaleMult, py+((self.y or 0)+(self.border or 0)-pCOffy)*scaleMult 
  self.drawW, self.drawH = (self.w or 0)*scaleMult, (self.h or 0)*scaleMult
      
  if self.l ~= nil then self.drawX = self.l[1](self,'left') end
  if self.t ~= nil then self.drawY = self.t[1](self,'top') end
  if self.r ~= nil then self.drawW = self.r[1](self,'right') end
  if self.b ~= nil then self.drawH = self.b[1](self,'bottom') end
  if self.minW ~= nil and self.drawW < self.minW then self.drawW = self.minW end
  if self.minH ~= nil and self.drawH < self.minH then self.drawH = self.minH end
  
  if self.img then 
    self.drawImg = scaleToDrawImg(self) -- adds _150 or _200 to name
    if self.imgIdx == nil then self.imgIdx = getImage(self.drawImg, self.iLocation, self.noIScales) end
    local iDw, iDh = gfx.getimgdim(self.imgIdx)
    if self.imgPink then iDw, iDh = iDw-2, iDh-2 end
    
    if self.measureImg == true then
      if self.iType ~= nil then
        if self.iType == 3 then iDw = iDw/3 end
        if self.iType == 'stack' then 
          self.measuredImgW, self.measuredImgH = iDw, iDh
          iDh = self.iFrameH 
        end
      end
      self.drawW, self.drawH = iDw, iDh
    else self.measuredImgW, self.measuredImgH = iDw, iDh 
    end
    
  end 
  
  local b = self.border or 0
  if self.flow then
    if self.flow == true then self.flow = getPreviousEl(self) or nil end -- auto set previous child as flow element
    if type(self.flow) == 'table' then
    
      local fx, fy = self.flow.drawX + self.flow.drawW + (self.x or 0) + b, self.flow.drawY
      if fx + b + self.drawW > px+pw then -- then flow to next row
        fx = (self.x or 0) + px + b
        fy = self.flow.drawY + self.flow.drawH + (self.y or 0) + b
      end
      self.drawX, self.drawY = fx, fy
      
    end 
  end
 
  resizeParent(self) -- checks if el has a resizeParent set, and if so does that
  
  if self.elAlign then
  
    if self.elAlign.x then
      local target = self.elAlign.x[1]
      if self.elAlign.x[1]=='parent' then target = self.parent end
      tx,tw = target.drawX or target.x, target.drawW or target.w
      if self.elAlign.x[2]=='centre' then self.drawX = tx+(tw/2)-(self.drawW/2)+ self.x end 
      if self.elAlign.x[2]=='right' then self.drawX = tx + tw - self.drawW + self.x end
    end
    
    if self.elAlign.y then
      local target = self.elAlign.y[1]
      if self.elAlign.y[1]=='parent' then target = self.parent end
      ty,th = target.drawY or target.y, target.drawH or target.h
      if self.elAlign.y[2]=='centre' then self.drawY = ty+(th/2)-(self.drawH/2)+self.y end
      if self.elAlign.y[2]=='bottom' then reaper.ShowConsoleMsg('element align bottom not done yet\n') end
    end

  end
  
  if self.clip==true then
    if (self.drawX + self.drawW +b > px+pw) or (self.drawY + self.drawH +b > py+ph) then 
      self.drawW = 0 
    end
  end
  
  --check final position, cull if outside parent
  if self.drawX > px+pw then -- fully to the right of parent
    self.drawW = 0 -- using zero width (instead of some kind of 'don't draw' state) so that dirtyCheck notices
  end

  return true
end




  --------- DRAW ----------
  
function El:draw(z,offsX,offsY)
  gfx.dest = (z or 1)
  local x,y,w,h = self.drawX or self.x or 0, self.drawY or self.y or 0, self.drawW or self.w or 0, self.drawH or self.h or 0
  x, y = x-offsX, y-offsY
  local col = self.drawCol or self.col or nil
  if col ~= nil then -- fill
    setCol(col)
    if self.shape ~= nil then
      if self.shape == 'circle' then gfx.circle(x+w/2,y+w/2,w/2,1,1) end
    else gfx.rect(x,y,w,h)
    end
  end
  if self.strokeCol ~= nil then -- stroke
    local c = fromRgbCol(self.strokeCol)
      gfx.set(c[1],c[2],c[3],c[4])
      if self.shape ~= nil then reaper.ShowConsoleMsg('non-rectangular strokes not done yet in El:draw\n') 
      else 
        gfx.line(x,y+h,x,y,0)
        gfx.line(x+1,y,x+w,y,0)
        gfx.line(x+w,y+1,x+w,y+h)
        gfx.line(x+1,y+h,x+w-1,y+h)
      end
  end
  if self.text ~= nil then
    if self.text.val ~=nil then self.text.str = self.text.val() end
    local tx,tw = x + textPadding, w - 2*textPadding
    text(self.text.str,tx,y,tw,h,self.text.align,self.text.col,((self.text.style + textScaleOffs) or 1),self.text.lineSpacing,self.text.vCenter,self.text.wrap)
  end
 
  if self.drawImg ~= nil then
 
    pinkXY, pinkWH = 0,0
    if self.imgPink then pinkXY, pinkWH = 1,2 end

    iDw, iDh =  self.drawW, self.drawH
    gfx.a = (self.img.a or 255) / 255
    if self.iType == 'stack' then 
      local iFrameHScaled = self.iFrameH * scaleMult  
      if self.iFrameC == nil then self.iFrameC = self.measuredImgH / iFrameHScaled end
      local frame = (self.iFrame or 0) * iFrameHScaled
      --reaper.ShowConsoleMsg('blit bitmap stack frame '..frame..' at '..x..'  '..y..'\n')
      gfx.blit(self.imgIdx, 1, 0, 0, frame, iDw, iFrameHScaled, x, y, w, iFrameHScaled)
      
    elseif self.iType == 'meterV' then
      local mColm = (self.meterColumn  or 1)*2 
      local stripW, displayVal = self.measuredImgW / 8, self.mValue or 0
      gfx.blit(self.imgIdx, 1, 0, stripW*(mColm-2), 0, stripW, self.measuredImgH, x, y, w, h)
      gfx.blit(self.imgIdx, 1, 0, stripW*(mColm-1), self.measuredImgH*(1-displayVal), stripW, self.measuredImgH*displayVal, x, y+(h*(1-displayVal)), w, h*displayVal)
    
    elseif self.iType == 'meterH' then
      local mRow = (self.meterColumn  or 1)*2 
      --reaper.ShowConsoleMsg('blit meter value '..self.mValue..' at '..x..'  '..y..', w: '..w..', h: '..h..', iDw: '..iDw..'\n')
      local stripH, displayVal = self.measuredImgH / 8, self.mValue or 0
      gfx.blit(self.imgIdx, 1, 0, 0, stripH*(mRow-2), self.measuredImgW, stripH, x, y, w, h)
      gfx.blit(self.imgIdx, 1, 0, 0, stripH*(mRow-1), self.measuredImgW*displayVal, stripH, x, y, w*displayVal, h)
      
    elseif self.iType == 'idxbgColorHack' then
      --gfx.blit(source[, scale, rotation, srcx, srcy, srcw, srch, destx, desty, destw, desth, rotxoffs, rotyoffs])
      gfx.blit(self.imgIdx, 1, 0, 1000, 4, 1, 1, x+16, y, w-16, h)
      
    elseif self.iType ~= nil then
      --reaper.ShowConsoleMsg('blit iType='..self.iType..' image='..self.img..' to z='..z..',   w='..w..',   h='..h..',   x='..x..', y='..y..' \n')
      gfx.blit(self.imgIdx, 1, 0, (self.iFrame or 0)*w +pinkXY, pinkXY, self.drawW, iDh, x, y, w, h)
    
    else gfx.blit(self.imgIdx, 1, 0, 0, 0, iDw, iDh, x, y, iDw, iDh + stretchH)
    end
    
  end
end




  --------- MOUSE ---------
  
function El:mouseOver()
  if self.mouseOverCol ~= nil then 
    self.drawCol = self.mouseOverCol
    self:addToDirtyZone()
  end
  if self.mouseOverCursor ~= nil then
    gfx.setcursor(429,1)
  end
  if self.img ~= nil then
    if self.iType ~= nil and self.iType == 3 then
      self.iFrame = 1
      self:addToDirtyZone()
    end
  end
end

function El:mouseAway()
  if self.mouseOverCol ~= nil then 
    self.drawCol = self.col
    self:addToDirtyZone()
  end
  if self.mouseOverCursor ~= nil then
    gfx.setcursor(1,1)
  end
  if self.img ~= nil then
    if self.iType ~= nil and self.iType == 3 then
      self.iFrame = 0
      self:addToDirtyZone()
    end
  end
end

function El:mouseDown(x,y)
  if self.img ~= nil then
    if self.iType ~= nil and self.iType == 3 then
      self.iFrame = 2
      self:addToDirtyZone()
    end
  end
  if self.onClick ~= nil and singleClick ~= true then
    singleClick = true
    self.onClick(self)
  end
  if self.onDrag then
    dX, dY = mouseDrag(self)
    self.onDrag(self, 'set', nil, dX, dY)
  end
end

function mouseDrag(self)
  if dragStart == nil then dragStart = {x=gfx.mouse_x, y=gfx.mouse_y} end
  local dX, dY = gfx.mouse_x - dragStart.x, gfx.mouse_y - dragStart.y
  
  local ctrl = gfx.mouse_cap&4
  if ctrl == 4 then -- ctrl
    if dragStart.fine ~= true then
      dragStart = {x=dragStart.x+dX, y=dragStart.y+dY}
      dragStart.fine = true
    end
    dX, dY = (gfx.mouse_x - dragStart.x)*0.25, (gfx.mouse_y - dragStart.y)*0.25
  end
  return dX, dY
end

function passToEl(self,...)
  --argsCheck = {...}
  self.El.action(self.El,...)
end

function El:doubleClick() 
  if self.onDoubleClick ~= nil then
    if type(self.onDoubleClick) == 'string' then
      if self.onDoubleClick == 'resetEl' then passToEl(self,'set','reset') end
      if self.onDoubleClick == 'reset' then self.action(self, 'set', 0) end
    else self.onDoubleClick(self)
    end
  end
end

function El:mouseWheel(wheel_amt)
  if self.action ~= nil then
    self.action(self,'set',wheel_amt)
  end
end





  --------- POPULATE ----------

--colDebug = true
--debugBuffers = true
setScale(1)
buttonStyles = {'alloy', 'soft', 'rounded', 'flat', 'dark_flat', 'transparent_flat', 'dark_planar', 'dark_rounded', 'default_5', 'grey_rounded', 
  'icon', 'icon_field', 'bordered', 'bordered_white', 'filled_flat', 'grey_field', 'grey_icon', 'traditional','reversed', 'sand', 'graphite', 'paper',
  'shadow', 'no_shadow'}
scaleFactor = 100

themeFolder = "Assembled_Theme_Resources"

Block:new({})
thinBox = El:new({block=1, z=1, x=0, y=0, w=800, h=800, col={50,50,50}, interactive=false,
  onGfxResize = function()
    thinBox.w, thinBox.h = gfx.w/scaleMult, gfx.h/scaleMult
    thinBox:addToDirtyZone(thinBox.z)
  end
  })
  
topBar = El:new({parent=thinBox, x=0, y=0, w=0, h=16, r={toEdge, thinBox, 'right'}, col={20,20,20} })

DPIdisplay = El:new({parent=topBar, x=0, w=200, y=0,h=16,
      text={str='', style=2, align=5, col={150,150,150}},
      onDpiChange = function() 
        DPIdisplay.text.str = 'THIS DISPLAY : '..math.floor(100*gfx.ext_retina)..'%        DRAW : '
        DPIdisplay:addToDirtyZone(DPIdisplay.z)
      end})
      
      
scale100 = El:new({parent=topBar, resizeParent='v', flow=DPIdisplay, z=1, x=0, y=0, w=40, h=16, col={20,20,20}, mouseOverCol={70,70,70}, 
      text={str='100%', style=2, align=5, col={150,150,150}},
      onClick = function() setScale(1) end })  
      
scale150 = El:new({parent=topBar, resizeParent='v', flow=scale100, z=1, x=2, y=0, w=40, h=16, col={20,20,20}, mouseOverCol={70,70,70}, 
      text={str='150%', style=2, align=5, col={150,150,150}},
      onClick = function() setScale(1.5) end })   
      
scale200 = El:new({parent=topBar, resizeParent='v', flow=scale150, z=1, x=2, y=0, w=40, h=16, col={20,20,20}, mouseOverCol={70,70,70}, 
      text={str='200%', style=2, align=5, col={150,150,150}},
      onClick = function() setScale(2) end })  
      
eraseButton = El:new({parent=topBar, resizeParent='v', flow=scale200, z=1, x=38, y=0, w=90, h=16, col={100,10,10}, mouseOverCol={255,70,70}, 
      text={str='ERASE THEME', style=2, align=5, col={0,0,0}},
      onClick = function() 
        suspendSwitchTheme = true
        local v = reaper.ShowMessageBox('Erase the theme file, then close the Theme Assembler?', 'ERASE THEME', 1)
        if v == 1 then 
          if os.remove(themes_path..'/Assembled_Theme.ReaperTheme') then
            reaper.OpenColorThemeFile(themes_path..'/Default.ReaperTheme')
            gfx.quit()
          end
        end
      end 
      }) 
      
      
      
orderThing = El:new({parent=thinBox, flow=topBar, x=10, y=10, w=400, h=70, col={60,60,60}, 
      onArrange = function()
      
        local str = themes_path..'/'..themeFolder..'/'..'rtconfig.txt'
        local rtConfigTable = {}
        local tableFilled = false
        
        for line in io.lines(str) do
          if string.find(line, 'then tcp') then
            table.insert(rtConfigTable, {line})
            tableFilled = true
          else
            if tableFilled == true then break end
          end
        end
        
        if tableFilled == true then
          for i,v in ipairs(rtConfigTable) do
          
            if i>=5 then -- apply new field assigments to the orderField elements
              for j, k in pairs({{'tcp.io',42},{'tcp.env',50},{'tcp_fx_group',44},{'tcp_pan_group',100},{'tcp.recmode',48},{'tcp_input_group',60}}) do
                if string.match(v[1],k[1]) then 
                  orderFields[i-4].field, orderFields[i-4].index, orderFields[i-4].w = k[1], i, k[2]
                end
              end
            end
            
          end
        end
        
      end
      })   

orderFields = {}
orderFields[1] = OrderField:new({rButton=true}) 
orderFields[2] = OrderField:new({flow=orderFields[1], rButton=true, lButton=true})
orderFields[3] = OrderField:new({flow=orderFields[2], rButton=true, lButton=true})
orderFields[4] = OrderField:new({flow=orderFields[3], rButton=true, lButton=true})
orderFields[5] = OrderField:new({flow=orderFields[4], rButton=true, lButton=true})
orderFields[6] = OrderField:new({flow=orderFields[5], lButton=true})      
      
      
      

globalBox = El:new({parent=thinBox, flow=orderThing, x=10, y=10, w=200, h=70, col={60,60,60}})

TargetButton:new({parent=globalBox, x=10, y=10, iType='themeFile', 
      buttonStyles = {'dark', 'def6', 'def5', 'cream', 'Classic'},
      styleProxyCols = {
                        {{51,51,51}, {72,72,72}, {45,79,71}, {106,106,106}},
                        {{51,51,51}, {129,137,137}, {19,189,153}, {69,69,69}},
                        {{51,51,51}, {156,164,164}, {117,214,145}, {190,192,192}},
                        {{217,217,204}, {171,162,143}, {102,102,77}, {255,255,255}},
                        {{240,240,240}, {210,205,210}, {118,59,142}, {223,114,26}}
                        },
      states = {
      off = {{'Theme File', 'Base .ReaperTheme file (code colors and font assignments)'}}
      }
  })

panelButton = TargetButton:new({parent=globalBox, x=126, y=10, iType='idxbgColorHack', title = 'tcp_idxbg',
      buttonStyles = {'100', '80', '60', '40', '20', '0', 'v5ish', 'panel'},
      states = {
      norm = {{'tcp_idxbg', 'unselected'}, {'mcp_idxbg', 'mixer unselected'}},
      sel = {{'tcp_idxbg_sel', 'selected'}, {'mcp_idxbg_sel', 'mixer selected'}}
      },
      onReaperChange = function()
        if projectCustCols == nil then getProjectCustCols() end
      end,
      nextFrameTime = 0,
      onFps = function()
        if projectCustCols ~= nil and #projectCustCols>0 and nowTime > panelButton.nextFrameTime then
          representCustCol = projectCustCols[math.random(1,#projectCustCols)]
          panelButton.col = representCustCol
          panelButton.nextFrameTime = nowTime + 2
          panelButton:addToDirtyZone()
        end
      end
  })
  
idxText = El:new({parent=globalBox, x=122, y=26, w=26, h=16, col={0,0,0,0}, mouseOverCol={255,255,255,100}, 
      text={str='88', style=2, align=5, col={255,255,255}},
      onArrange = function()
        local n = getRtconfigVar('textOverCCbrightness') 
        idxText.text.col = {n,n,n}
      end,
      onClick = function() 
        local val = getRtconfigVar('textOverCCbrightness') 
        local r, v = reaper.GetUserInputs('Track Number Text', 1, 'Brightness (0-255)',val)
        local n = math.Clamp(tonumber(v),0,255)
        setRtconfigVar('textOverCCbrightness', n)
        idxText.text.col = {n,n,n}
      end
  }) 
  
  
  
  
---- Transport Box ----

transportBox = El:new({parent=thinBox, flow=globalBox, x=10, y=20, w=230, h=70, col={50,80,50}, interactive=false})

TargetButton:new({parent=transportBox, x=10, y=10, title = 'transport_home',
      states = {
      off = {{'transport_home', 'Home'}, {'transport_previous','Previous'}}
      }
  })
  
TargetButton:new({parent=transportBox, x=39, y=10, title = 'transport_end',
      states = {
      off = {{'transport_end', 'End'}, {'transport_next','Next'}}
      }
  })
  
TargetButton:new({parent=transportBox, x=66, y=7, title = 'transport_record',
      states = {
      off = {{'transport_record', 'Record'}, {'transport_record_item','Record Item'}, {'transport_record_loop','Record Loop'}},
      on = {{'transport_record_on', 'Record On'}, {'transport_record_item_on','Record Item On'}, {'transport_record_loop_on','Record Loop On'}}
      }
  })
  
TargetButton:new({parent=transportBox, x=100, y=7, title = 'transport_play',
      states = {
      off = {{'transport_play', 'Play'}, {'transport_play_sync','Play Sync'}},
      on = {{'transport_play_on', 'Play On'}, {'transport_play_sync_on','Play Sync On'}}
      }
  })
  
TargetButton:new({parent=transportBox, x=133, y=10, title = 'transport_repeat_off',
      states = {off = {{'transport_repeat_off', 'Repeat'}}, on = {{'transport_repeat_on', 'Repeat On'}} }
  })
  
TargetButton:new({parent=transportBox, x=160, y=10, title = 'transport_stop',
      states = {off = {{'transport_stop', 'Stop'}} }
  })
  
TargetButton:new({parent=transportBox, x=189, y=10, title = 'transport_pause',
      states = {off = {{'transport_pause', 'Pause'}}, on = {{'transport_pause_on', 'Pause On'}} }
  })
  
  


---- TCP Box ----

tcpBox = El:new({parent=thinBox, flow=transportBox, x=10, y=10, w=88, h=70, col={80,50,50}, interactive=false})

knobBg = El:new({parent=tcpBox, x=9, y=9, img='track_recarm_off_ol', iType=3, imgPink=true})

TargetButton:new({parent=tcpBox, x=10, y=10, title = 'track_recarm_off', imgPinkStates = {'shadow'},
      states = {
      off = {{'track_recarm_off', 'Record Arm'}, {'track_recarm_auto','Automatic Arm'}, {'mcp_recarm_off', 'Mixer Record Arm'}, {'mcp_recarm_auto','Mixer Automatic Arm'}},
      on = {{'track_recarm_on','Record Arm On'}, {'track_recarm_auto_on','Automatic Arm On'}, {'track_recarm_norec', 'Record Disabled'}, {'track_recarm_auto_norec','Automatic Disabled'}, 
            {'mcp_recarm_on','Mixer Record Arm On'}, {'mcp_recarm_auto_on','Mixer Automatic Arm On'}, {'mcp_recarm_norec', 'Mixer Record Disabled'}, {'mcp_recarm_auto_norec','Mixer Automatic Disabled'}},
      shadow = {{'track_recarm_off_ol', 'Record Arm Shadow'}, {'track_recarm_auto_ol','Automatic Arm Shadow'}, {'mcp_recarm_off_ol', 'Mixer Record Arm Shadow'}, {'mcp_recarm_auto_ol','Mixer Automatic Arm Shadow'}}
      }
  })
 
TargetButton:new({parent=tcpBox, x=10, y=40, title = 'mcp_solo_off_ol', imgPink = true, imgPinkStates = {'off','on','defeat'},
      states = {
      off = {{'mcp_solo_off_ol', 'Mixer Solo'},{'track_solo_off_ol', 'Track Solo'}}, 
      on = {{'mcp_solo_on_ol', 'Mixer Solo On'},{'track_solo_on_ol', 'Track Solo On'}},
      defeat = {{'mcp_solodefeat_on_ol','Mixer Solo Defeat'}, {'tcp_solodefeat_on_ol','Track Solo Defeat'}}
      }
  })

TargetButton:new({parent=tcpBox, x=30, y=40, title = 'track_mute_off_ol', imgPink = true, imgPinkStates = {'off'},
      states = {
      off = {{'track_mute_off_ol', 'Mute'}},
      on = {{'track_mute_on_ol', 'Mute On'}}
      }
  })
  
TargetButton:new({parent=tcpBox, x=60, y=10, title = 'track_phase_norm_ol',
      states = {
      off = {{'track_phase_norm_ol', 'Phase Normal'}},
      on = {{'track_phase_inv_ol', 'Phase Inverted'}}
      }
  }) 
  
  
knobBg = El:new({parent=tcpBox, x=31, y=8, img='tcp_vol_knob_small', iType=1, imgPink=true})

TargetButton:new({parent=knobBg, x=2, y=2, z=1, title = 'tcp_vol_knob_stack', iType = 'stack', iTypeStates={stack='stack',knob=1}, iFrame = 1,
      onFps = cycleBitmapStack, imgPinkStates = {'knob'},
      buttonStyles = {'flower', 'radar', 'segment', 'segment_wide', 'arc', 'black_line', 'white_line', 'green_line', 'black_dot', 'white_dot',
                      'swipe_green', 'swipe_grey', 'sand', 'graphite', 'paper',
                      'alloy', 'dark_alloy', 'dark_flat', 'dark_rounded', 'grey', 'none', 'soft', 'light_post', 'dark_post', 'flat_post'},
      states = {
      stack = {{'tcp_vol_knob_stack', 'Volume Knob Pointer'}}, 
      knob = {{'tcp_vol_knob_small', 'Volume Knob'}}
      }
  })
  
TargetButton:new({parent=tcpBox, fieldParent='tcp.io', x=4, y=6, title = 'track_io', imgPink=true, imgPinkStates = {'off', 'shadow'}, 
      states = {
      off = {{'track_io', 'Routing M'}, {'track_io_dis', 'Routing'}, {'mcp_io', 'Mixer Routing M'}, {'mcp_io_dis', 'Mixer Routing'}, {'track_io_r','Routing MR'}, {'mcp_io_r','Mixer Routing MR'}, {'track_io_r_dis', 'Routing R'}, {'mcp_io_r_dis', 'Mixer Routing R'}, 
             {'track_io_s','Routing MS'}, {'mcp_io_s','Mixer Routing MS'}, {'track_io_s_dis','Routing S'}, {'mcp_io_s_dis','Mixer Routing S'}, {'track_io_s_r','Routing MSR'}, {'mcp_io_s_r','Mixer Routing MSR'}, 
             {'track_io_s_r_dis','Routing SR'}, {'mcp_io_s_r_dis','Mixer Routing SR'}},
      shadow = {{'track_io_ol', 'Routing M Shadow'}, {'track_io_dis_ol', 'Routing Shadow'}, {'track_io_r_ol','Routing MR Shadow'}, {'track_io_r_dis_ol', 'Routing R Shadow'}, 
             {'track_io_s_ol','Routing MS Shadow'}, {'track_io_s_dis_ol','Routing S Shadow'}, {'track_io_s_r_ol','Routing MSR Shadow'}, {'track_io_s_r_dis_ol','Routing SR Shadow'}}
      }
  })

TargetButton:new({parent=tcpBox, fieldParent='tcp.env', x=4, y=6, title = 'track_env', imgPink=true, imgPinkStates = {'off', 'shadow'},
      states = {
      off = {{'track_env', 'Trim'}, {'track_env_latch', 'Latch'}, {'track_env_preview', 'Preview'}, {'track_env_read', 'Read'}, {'track_env_touch', 'Touch'}, 
            {'track_env_write', 'Write'}},
      shadow = {{'track_env_ol', 'Trim Shadow'}, {'track_env_latch_ol', 'Latch Shadow'}, {'track_env_preview_ol', 'Preview Shadow'}, 
          {'track_env_read_ol', 'Read Shadow'}, {'track_env_touch_ol', 'Touch Shadow'}, {'track_env_write_ol', 'Write Shadow'}}
      }
  })  
  
TargetButton:new({parent=tcpBox, fieldParent='tcp_fx_group', x=4, y=6, title = 'track_fx_empty', imgPink=true, imgPinkStates = {'fx', 'shadow'},
      states = {
      fx = {{'track_fx_empty', 'Empty'},{'track_fx_norm', 'Normal'},{'track_fx_dis', 'Disabled'}},
      shadow = {{'track_fx_empty_ol', 'Empty Shadow'},{'track_fx_norm_ol', 'Normal Shadow'},{'track_fx_dis_ol', 'Disabled Shadow'}}
      }
  })

TargetButton:new({parent=tcpBox, fieldParent='tcp_fx_group', x=24, y=6, title = 'track_fxempty_h', imgPink=true, imgPinkStates = {'fx', 'shadow'},
      states = {
      fx = {{'track_fxempty_h', 'Empty'},{'track_fxon_h', 'Normal'},{'track_fxoff_h', 'Disabled'}},
      shadow = {{'track_fxempty_h_ol', 'Empty Shadow'},{'track_fxon_h_ol', 'Normal Shadow'},{'track_fxoff_h_ol', 'Disabled Shadow'}}
      }
  })
  
TargetButton:new({parent=tcpBox, fieldParent='tcp.recmode', x=4, y=6, title = 'track_recmode_in', imgPink=true, imgPinkStates = {'input', 'shadow'},
      states = {
      input = {{'track_recmode_in', 'In'},{'track_recmode_out', 'Out'},{'track_recmode_off', 'Off'}},
      shadow = {{'track_recmode_in_ol', 'Input'},{'track_recmode_out_ol', 'Output'},{'track_recmode_off_ol', 'Disabled'}}
      }
  })
  
TargetButton:new({parent=tcpBox, fieldParent='tcp_input_group', x=4, y=6, title = 'track_fx_in_empty', imgPink=true, imgPinkStates = {'fx', 'shadow'},
      states = {
      fx = {{'track_fx_in_empty', 'Empty'},{'track_fx_in_norm', 'Normal'}},
      shadow =  {{'track_fx_in_empty_ol', 'Empty Shadow'}}
      }
  }) 
  
TargetButton:new({parent=tcpBox, fieldParent='tcp_input_group', x=34, y=6, title = 'tcp_recinput', iType = 'bg', imgPink=true, imgPinkStates = {'input'},
      states = {
      input = {{'tcp_recinput', 'Input'}} 
      }
  })
  
tcpPanKnobBg = El:new({parent=tcpBox, fieldParent='tcp_pan_group', x=4, y=4, img='tcp_pan_knob_small', iType=1, imgPink=true})

TargetButton:new({parent=tcpPanKnobBg, x=1, y=1, z=1, title = 'tcp_pan_knob_stack', iType = 'stack', iTypeStates={stack='stack',knob=1}, iFrame = 1,
      onFps = cycleBitmapStack, imgPinkStates = {'knob'},
      buttonStyles = {'trails', 'black_dot', 'green_dot', 'grey_dot', 'white_dot', 'black_line', 'green_line', 'grey_line', 'white_line',
                      'black_swipe', 'green_swipe', 'grey_swipe', 'white_swipe', 'swap_swipe',
                      'pale', 'medium', 'alloy', 'dark_alloy', 'dark_flat', 'dark_rounded', 'grey', 'none', 'soft', 'dark_field', 'sand',  'graphite', 'paper'},
      states = {
      stack = {{'tcp_pan_knob_stack', 'Pan Knob Pointer'}}, 
      knob = {{'tcp_pan_knob_small', 'Pan Knob'}}
      }
  })
  
tcpWidthKnobBg = El:new({parent=tcpBox, fieldParent='tcp_pan_group', x=28, y=4, img='tcp_width_knob_small', iType=1, imgPink=true})

TargetButton:new({parent=tcpWidthKnobBg, x=1, y=1, z=1, title = 'tcp_wid_knob_stack', iType = 'stack', iTypeStates={stack='stack',knob=1}, iFrame = 1,
      onFps = cycleBitmapStack, imgPinkStates = {'knob'},
      buttonStyles = {'trails', 'black_dot', 'green_dot', 'grey_dot', 'white_dot', 'black_line', 'green_line', 'grey_line', 'white_line',
                      'black_swipe', 'green_swipe', 'grey_swipe', 'white_swipe', 'swap_swipe',
                      'pale', 'medium', 'alloy', 'dark_alloy', 'dark_flat', 'dark_rounded', 'grey', 'none', 'soft', 'dark_field', 'sand',  'graphite', 'paper'},
      states = {
      stack = {{'tcp_wid_knob_stack', 'Width Knob Pointer'}}, 
      knob = {{'tcp_width_knob_small', 'Width Knob'}}
      }
  }) 
  
TargetButton:new({parent=tcpBox, fieldParent='tcp_pan_group', x=54, y=6, title='tcp_panbg', iType=1, imgPink=true, imgPinkStates = {'bg'},
      buttonStyles = {'trans_flat','dark_flat','dark_rounded','trans_rounded','icon_field'},
      states = {
      bg = {{'tcp_panbg', 'Pan Background'}}
      }
  })
  
TargetButton:new({parent=tcpBox, fieldParent='tcp_pan_group', x=58, y=1, title = 'tcp_panthumb', iType=1, imgPink=true, imgPinkStates = {'tcp'},
      buttonStyles = {'soft', 'dark', 'trans_flat', 'dark_flat', 'alloy', 'dark_alloy'},
      states = {tcp = {{'tcp_panthumb', 'Pan Fader'}} }
  })  
  
TargetButton:new({parent=tcpBox, fieldParent='tcp_pan_group', x=54, y=16, title='tcp_widthbg', iType=1, imgPink=true, imgPinkStates = {'bg'},
      buttonStyles = {'trans_flat','dark_flat','dark_rounded','trans_rounded','icon_field'},
      states = {
      bg = {{'tcp_widthbg', 'Width Background'}}
      }
  })
  
TargetButton:new({parent=tcpBox, fieldParent='tcp_pan_group', x=80, y=11, title = 'tcp_widththumb', iType=1, imgPink=true, imgPinkStates = {'tcp'},
      buttonStyles = {'soft', 'dark', 'trans_flat', 'dark_flat', 'alloy', 'dark_alloy'},
      states = {tcp = {{'tcp_widththumb', 'Width Fader'}} }
  })  
  
  
  
---- MCP Box ----  
  
mcpBox = El:new({parent=thinBox, flow=tcpBox, x=10, y=10, w=320, h=70, col={50,50,80}, interactive=false})  
  
TargetButton:new({parent=mcpBox, x=10, y=10, buttonH=70, title = 'mcp_volthumb', iType=1, imgPink=true, imgPinkStates = {'mcp', 'tcp'}, 
      buttonStyles = {'alloy', 'soft', 'soft_dark', 'dark', 'dark_&_green', 'grey', 'Imperial', 'Imperial_Dark', 'sand', 'graphite', 'paper'},
      states = {mcp = {{'mcp_volthumb', 'Mixer Fader'}}, tcp = {{'tcp_volthumb', 'Track Fader'}} }
  })  
  
TargetButton:new({parent=mcpBox, x=40, y=10, title = 'mcp_recinput', iType = 'bg', imgPink=true, imgPinkStates = {'input'},
      states = {
      input = {{'mcp_recinput', 'Mixer Input'}} 
      }
  })
  
inputText = El:new({parent=mcpBox, x=34, y=10, w=18, h=16, col={0,0,0,0}, mouseOverCol={255,255,255,100}, 
      text={str='8', style=2, align=5, col={255,255,255}},
      onArrange = function()
        local n = getRtconfigVar('inputTextBrightness') 
        inputText.text.col = {n,n,n}
      end,
      onClick = function() 
        local val = getRtconfigVar('inputTextBrightness') 
        local r, v = reaper.GetUserInputs('Input Text', 1, 'Brightness (0-255)',val)
        local n = math.Clamp(tonumber(v),0,255)
        setRtconfigVar('inputTextBrightness', n)
        inputText.text.col = {n,n,n}
      end
  })

mcpFx = TargetButton:new({parent=mcpBox, x=70, y=10, title = 'mcp_fx_empty', imgPink=true, imgPinkStates = {'fx'},
      states = {
      fx = {{'mcp_fx_empty', 'Empty'},{'mcp_fx_norm', 'Normal'},{'mcp_fx_dis', 'Disabled'}}
      }
  })
  
TargetButton:new({parent=mcpBox, flow = mcpFx, x=0, y=0, title = 'track_fxempty_v', imgPink=true, imgPinkStates = {'fx'},
      states = {
      fx = {{'track_fxempty_v', 'Empty'},{'track_fxon_v', 'Normal'},{'track_fxoff_v', 'Disabled'}}
      }
  })

TargetButton:new({parent=mcpBox, x=40, y=40, title = 'mcp_fx_in_empty_ol', imgPink=true, imgPinkStates = {'fx'},
      states = {
      fx = {{'mcp_fx_in_empty_ol', 'Empty'},{'mcp_fx_in_norm_ol', 'Normal'}}
      }
  })
  
TargetButton:new({parent=mcpBox, x=130, y=10, title = 'mcp_recmode_in', imgPink=true, imgPinkStates = {'input', 'shadow'},
      states = {
      input = {{'mcp_recmode_in', 'In'},{'mcp_recmode_out', 'Out'},{'mcp_recmode_off', 'Off'}},
      shadow = {{'mcp_recmode_in_ol', 'Borked'},{'mcp_recmode_out_ol', 'Borked'},{'mcp_recmode_off_ol', 'Bastard'}}
      }
  })  
  
TargetButton:new({parent=mcpBox, x=196, y=10, z=1, w=12, h=50, buttonH=200, title = 'meter_strip_v', iType = 'meterV', noIScales = true, onFps = fakeMeterVals,
      buttonStyles = {'v6', 'NoArm', 'gradient', 'green_to_red', 'g2r-gradient', 'discharge', 'discharge+', 'grey', '0-6-12', '0-6-18', '0-12-18', '0-12-24'},
      states = {
      meter_v = {{'meter_strip_v', 'Vertical Meter'}}
      }
  })

TargetButton:new({parent=mcpBox, x=130, y=40, z=1, w=50, h=12, buttonW=200, title = 'meter_strip_h', iType = 'meterH', noIScales = true, onFps = fakeMeterVals,
      buttonStyles = {'v6', 'NoArm', 'gradient', 'green_to_red', 'g2r-gradient', 'discharge', 'discharge+', 'grey', '0-6-12', '0-6-18', '0-12-18', '0-12-24'},
      states = {
      meter_v = {{'meter_strip_h', 'Horizontal Meter'}}
      }
  })
  
mcpPanKnobBg = El:new({parent=mcpBox, x=230, y=40, img='mcp_pan_knob_small', iType=1, imgPink=true})

TargetButton:new({parent=mcpPanKnobBg, x=1, y=1, z=1, title = 'mcp_pan_knob_stack', iType = 'stack', iTypeStates={stack='stack',knob=1}, iFrame = 1,
      onFps = cycleBitmapStack, imgPinkStates = {'knob'},
      buttonStyles = {'trails', 'black_dot', 'green_dot', 'grey_dot', 'white_dot', 'black_line', 'green_line', 'grey_line', 'white_line',
                      'black_swipe', 'green_swipe', 'grey_swipe', 'white_swipe', 'swap_swipe',
                      'pale', 'medium', 'alloy', 'dark_alloy', 'dark_flat', 'dark_rounded', 'grey', 'none', 'soft', 'dark_field', 'sand',  'graphite', 'paper'},
      states = {
      stack = {{'mcp_pan_knob_stack', 'Mixer Pan Knob Pointer'}}, 
      knob = {{'mcp_pan_knob_small', 'Mixer Pan Knob'}}
      }
  })
  
mcpWidthKnobBg = El:new({parent=mcpBox, x=254, y=40, img='mcp_width_knob_small', iType=1, imgPink=true})

TargetButton:new({parent=mcpWidthKnobBg, x=1, y=1, z=1, title = 'mcp_wid_knob_stack', iType = 'stack', iTypeStates={stack='stack',knob=1}, iFrame = 1,
      onFps = cycleBitmapStack, imgPinkStates = {'knob'},
      buttonStyles = {'trails', 'black_dot', 'green_dot', 'grey_dot', 'white_dot', 'black_line', 'green_line', 'grey_line', 'white_line',
                      'black_swipe', 'green_swipe', 'grey_swipe', 'white_swipe', 'swap_swipe',
                      'pale', 'medium', 'alloy', 'dark_alloy', 'dark_flat', 'dark_rounded', 'grey', 'none', 'soft', 'dark_field', 'sand',  'graphite', 'paper'},
      states = {
      stack = {{'mcp_wid_knob_stack', 'Mixer Width Knob Pointer'}}, 
      knob = {{'mcp_width_knob_small', 'Mixer Width Knob'}}
      }
  })
  
TargetButton:new({parent=mcpBox, x=216, y=8, title='mcp_panbg', iType=1, imgPink=true, imgPinkStates = {'bg'},
      buttonStyles = {'trans_flat','dark_flat','dark_rounded','trans_rounded','icon_field'},
      states = {
      bg = {{'mcp_panbg', 'Mixer Pan Background'}}
      }
  })
  
TargetButton:new({parent=mcpBox, x=226, y=3, title = 'mcp_panthumb', iType=1, imgPink=true, imgPinkStates = {'mcp'},
      buttonStyles = {'soft', 'dark', 'trans_flat', 'dark_flat', 'alloy', 'dark_alloy'},
      states = {mcp = {{'mcp_panthumb', 'Pan Fader'}} }
  })  
  
TargetButton:new({parent=mcpBox, x=216, y=18, title='mcp_widthbg', iType=1, imgPink=true, imgPinkStates = {'bg'},
      buttonStyles = {'trans_flat','dark_flat','dark_rounded','trans_rounded','icon_field'},
      states = {
      bg = {{'mcp_widthbg', 'Mixer Width Background'}}
      }
  })
  
TargetButton:new({parent=mcpBox, x=266, y=13, title = 'mcp_widththumb', iType=1, imgPink=true, imgPinkStates = {'mcp'},
      buttonStyles = {'soft', 'dark', 'trans_flat', 'dark_flat', 'alloy', 'dark_alloy'},
      states = {mcp = {{'mcp_widththumb', 'Width Fader'}} }
  })  

TargetButton:new({parent=mcpBox, x=292, y=6, title = 'mcp_env', imgPink=true, imgPinkStates = {'off'},
      states = {
      off = {{'mcp_env', 'Mixer Trim'}, {'mcp_env_latch', 'Mixer Latch'}, {'mcp_env_preview', 'Mixer Preview'}, {'mcp_env_read', 'Mixer Read'}, {'mcp_env_touch', 'Mixer Touch'}, {'mcp_env_write', 'Mixer Write'}},
      }
  })
       


  --------- RUNLOOP ----------
 
fps = 24 -- hohoho
lastchgidx = 0
mouseXold = 0
mouseYold = 0
mouseWheelAccum = 0
trackCountOld = 0
dirtyZones ={}
bufIdx = {}
themeCheck()
gfxWold, gfxHold = gfx.w, gfx.h


function runloop()
  c=gfx.getchar()
  themeCheck()
 
  -- mouse stuff
  local isCap = (gfx.mouse_cap&1)
  
  if gfx.mouse_x ~= mouseXold or gfx.mouse_y ~= mouseYold or (firstClick ~= nil and last_click_time ~= nil and last_click_time+.25 < nowTime) then
    firstClick = nil
  end
  
  if gfx.mouse_x ~= mouseXold or gfx.mouse_y ~= mouseYold or isCap ~= mouseCapOld or gfx.mouse_wheel ~= 0 then
  
    local wheel_amt = 0
    if gfx.mouse_wheel ~= 0 then
      mouseWheelAccum = mouseWheelAccum + gfx.mouse_wheel
      gfx.mouse_wheel = 0
      wheel_amt = math.floor(mouseWheelAccum / 120 + 0.5)
      if wheel_amt ~= 0 then mouseWheelAccum = 0 end
    end
    
    local hit = nil
    
    for b in ipairs(els) do -- iterate blocks
      for z = #els[b],1,-1 do -- iterate z backwards
        if els[b][z] ~= nil then
          for j,k in pairs(els[b][z]) do
            
            local x, y, w, h = k.drawX or k.x or 0, k.drawY or k.y or 0, k.drawW or k.w or 0, k.drawH or k.h or 0
            if k.hitBox then x, y, w, h = x + k.hitBox[1], y + k.hitBox[2], k.hitBox[3], k.hitBox[4] end
            if k.interactive ~= false and gfx.mouse_x > x and gfx.mouse_x < x+w and gfx.mouse_y > y and gfx.mouse_y < y+h ~= false then
              hit = k
            end
  
          end
        end
      
      end
    end
    
    if isCap == 0 or mouseCapOld == 0 then
      if activeMouseElement ~= nil and activeMouseElement ~= hit then
        activeMouseElement:mouseAway()
        singleClick = nil
      end
      activeMouseElement = hit
    end
    
    if isCap == 0 and mouseCapOld == 1 then -- mouse-up, reset stuff
      dragStart, singleClick = nil, nil
    end
    
    if activeMouseElement ~= nil then
      if isCap == 0 or mouseCapOld == 0 then
        activeMouseElement:mouseOver()
      end
      if wheel_amt ~= 0 then       
        activeMouseElement:mouseWheel(wheel_amt)
      end
       
      if isCap == 1 then -- mouse down
        activeMouseElement:mouseDown()
         
         local x,y = gfx.mouse_x,gfx.mouse_y
         if firstClick == nil or last_click_time == nil then 
           firstClick = {gfx.mouse_x,gfx.mouse_y}
           last_click_time = nowTime
         else if nowTime < last_click_time+.25 and math.abs((x-firstClick[1])*(x-firstClick[1]) + (y- firstClick[2])*(y- firstClick[2])) < 4 then 
           activeMouseElement:doubleClick() 
           firstClick = nil
           else
             firstClick = nil
           end 
         end
         
      end
    end
    
    mouseXold, mouseYold, mouseCapOld = gfx.mouse_x, gfx.mouse_y, isCap
  end
  

  
  
  -- changes every FPS
  nowTime = reaper.time_precise()
  if (nextTime == nil or nowTime > nextTime) then -- do onFrame updates
  
    for b in ipairs(els) do -- iterate blocks
      for z in ipairs(els[b]) do -- iterate z
        if els[b][z] ~= nil then
          
          for j,k in pairs(els[b][z]) do
            if k.onFps and k.w ~= 0 then
              k.onFps(k)
            end
          end
          
        end
      end
    end

    nextTime = nowTime + (1/fps)
    lastTime = nowTime
  end
  
  -- changes from Reaper
  chgidx = reaper.GetProjectStateChangeCount(0)
  if chgidx ~= lastchgidx or doReaperGet == true then
    --reaper.ShowConsoleMsg('responding to changes from Reaper\n')
    for b in ipairs(els) do -- iterate blocks
      for z in ipairs(els[b]) do -- iterate z
        if els[z] ~= nil then
          for j,k in pairs(els[b][z]) do

            if k.onReaperChange then k.onReaperChange(k,'get') end
          end
        end
      end
    end
    
    lastchgidx = chgidx
    doReaperGet = false
    doArrange = true 
  end
 
  if gfxWold ~= gfx.w or gfxHold ~= gfx.h then
    for b in ipairs(els) do -- iterate blocks 
      for z in ipairs(els[b]) do -- iterate z
        if els[b][z] ~= nil then
          for j,k in pairs(els[b][z]) do
            
            if k.onGfxResize then
              k.onGfxResize(k)
              --reaper.ShowConsoleMsg('should I resize a block?\n')
            end
            
            doArrange = true
          end
        end
      end
    end
    gfxWold, gfxHold = gfx.w, gfx.h
  end
  
  
  -- change in screen DPI
  if gfx.ext_retina ~= ext_retinaOld or ext_retinaOld == nil then
    --reaper.ShowConsoleMsg('DPI changed or first time\n')
    
    for b in ipairs(els) do -- iterate blocks
      for z in ipairs(els[b]) do -- iterate z
        if els[b][z] ~= nil then
          for j,k in pairs(els[b][z]) do -- iterate elements
            if k.onDpiChange and k.w ~= 0 then
              k.onDpiChange(k)
              doArrange = true
            end
          end
        end
      end
    end
    
    local nScale = 1
    if gfx.ext_retina > 1.33 then nScale = 1.5 end
    if gfx.ext_retina > 1.66 then nScale = 2 end
    setScale(nScale)
    ext_retinaOld = gfx.ext_retina
  end
  
  
 
  -- ARRANGE --
  
  if doArrange == true then -- do Arrange
    --reaper.ShowConsoleMsg('do Arrange\n')
    for b in ipairs(els) do -- iterate blocks 
      for z in ipairs(els[b]) do
        if els[b][z] ~= nil then
          for j,k in pairs(els[b][z]) do 
            k:dirtyXywhCheck(b,z) -- dirtyXywhCheck stores the old xywh, arranges the element, then adds to dirtyZone if it is dirty
          end
        end
      end
    end
    
    --doArrange = false
  end




  -- DRAW --  
  
  if doDraw == true then

    for b in ipairs(els) do -- iterate blocks
      if bufIdx[b] == nil then bufIdx[b] = {} end
      local thisBx, thisBy = els[b][1][1].drawX or 0, els[b][1][1].drawY or 0 
      local prevZDirtyZone = {} -- used to copy dZ from a Z to the next Z, reseting for each block
      
      for z in ipairs(els[b]) do -- iterate z
        
        if prevZDirtyZone.x or (dirtyZones[b][z] ~= nil and dirtyZones[b][z].x2 ~= nil) then -- there is a dirtyZone

          if prevZDirtyZone.x then -- prevZDirtyZone is not empty
            if dirtyZones[b] == nil then dirtyZones[b] = {} end                  --this is a lot like the addToDirtyZone function, maybe that could do this
            if dirtyZones[b][z] == nil then dirtyZones[b][z] = {x1={},y1={},x2={},y2={}} end
            dirtyZones[b][z].x1[#dirtyZones[b][z].x1+1] = prevZDirtyZone.x
            dirtyZones[b][z].y1[#dirtyZones[b][z].y1+1] = prevZDirtyZone.y
            dirtyZones[b][z].x2[#dirtyZones[b][z].x2+1] = prevZDirtyZone.x + prevZDirtyZone.w
            dirtyZones[b][z].y2[#dirtyZones[b][z].y2+1] = prevZDirtyZone.y + prevZDirtyZone.h
            prevZDirtyZone = {}
          end
        
          -- FIND THE EXTENTS OF THE DIRTYZONE OF THIS Z --
          local dx1, dy1, dx2, dy2 = nil, nil, nil, nil
          for i=1, #dirtyZones[b][z].x1 do
            if dx1 == nil then dx1 = dirtyZones[b][z].x1[1] end
            if dirtyZones[b][z].x1[i] < dx1 then dx1 = dirtyZones[b][z].x1[i] end
          end
          for i=1, #dirtyZones[b][z].y1 do
            if dy1 == nil then dy1 = dirtyZones[b][z].y1[1] end
            if dirtyZones[b][z].y1[i] < dy1 then dy1 = dirtyZones[b][z].y1[i] end
          end
          for i=1, #dirtyZones[b][z].x2 do
            if dx2 == nil then dx2 = dirtyZones[b][z].x2[1] end
            if dirtyZones[b][z].x2[i] > dx2 then dx2 = dirtyZones[b][z].x2[i] end
          end
          for i=1, #dirtyZones[b][z].x2 do
            if dy2 == nil then dy2 = dirtyZones[b][z].y2[1] end
            if dirtyZones[b][z].y2[i] > dy2 then dy2 = dirtyZones[b][z].y2[i] end
          end
          local dx, dy, dw, dh = dx1, dy1, dx2 - dx1, dy2 - dy1 --dx dy dw dh are the coordinates of the dirtyZone, in screen coordinates
          prevZDirtyZone = {x=dx, y=dy, w=dw, h=dh}
          
          if bufIdx[b][z] == nil then -- this buffer doesn't exist
            if nextBufIdx == nil then nextBufIdx = 1 end
            bufIdx[b][z] = nextBufIdx
            gfx.setimgdim(nextBufIdx, dx-thisBx+dw, dy-thisBy+dh)
            nextBufIdx = nextBufIdx + 1
          else
            local iw, ih = gfx.getimgdim(bufIdx[b][z])
            if iw>0 and ((iw < dx-thisBx+dw) or (ih < dy-thisBy+dh)) then -- check whether this z buffer exists but is too small
              gfx.setimgdim(bufIdx[b][z], dx-thisBx+dw, dy-thisBy+dh) -- fix the buffer size
            end
          end
          
          gfx.dest = 9                                  -- the temp buffer.
          gfx.mode = 2
          gfx.setimgdim(9,dw,dh)                       -- prepare the new temp buffer 
          gfx.set(0)
          gfx.rect(0,0,dw,dh)                          -- erase it
        
          if bufIdx[b][z-1] then -- there is a buffer for the previous z
            gfx.blit(bufIdx[b][z-1],1,0,dx-thisBx,dy-thisBy,dw,dh,0,0,dw,dh) -- blit dZ from clean buffer of previous z BOTH IN BUFFER COORDS NOT SCREEN COORDS!!
          end
          
          for j,k in pairs(els[b][z]) do               -- iterate Els to draw to the buffer
            gfx.mode = 0
            k:drawToBuffer(9,dx,dy,dw,dh)     -- drawToBuffer will decide whether it needs drawing, and if so then self:draw to the temp buffer
          end
           
          -- VISIBLE BUFFERS --
          if debugBuffers == true then
            local border = 0
            --gfx.muladdrect(dx-thisBx+border,dy-thisBy+border,dw-border-border,dh-border-border,1,1,1,1,math.random()/3,math.random()/3,math.random()/3,0)
            gfx.muladdrect(dx,dy,dw,dh,1,1,1,1,math.random()/3,math.random()/3,math.random()/3,0)
          end
          
          --draw the temp buffer to the clean buffer
          gfx.mode = 2
          gfx.dest = bufIdx[b][z]
          gfx.blit(9,1,0, 0, 0, gfx.w, gfx.h, dx-thisBx, dy-thisBy, gfx.w, gfx.h)
          
        else --reaper.ShowConsoleMsg('block'..b..' layer'..z..' was CLEAN\n') 
        end -- end of this dirty z

        dirtyZones[b][z] = nil
      
      end -- end iterating z
    end -- end iterating blocks
    doDraw = false 

  end
  
  
  
  --COMP--
  for b in ipairs(els) do -- iterate blocks
    gfx.dest = -1 
    local blockW, blockH = els[b][1][1].drawW,els[b][1][1].drawH -- read draw sizes from the first el
    gfx.blit(bufIdx[b][#bufIdx[b]],1,0, 0, 0, blockW, blockH, els[b][1][1].drawX, els[b][1][1].drawY, blockW, blockH) 
  end
  
  if c>48 and c<59 then
    debugDrawZ = math.floor(c - 48)
  end
  
  if debugDrawZ ~= nil then 
    gfx.a, gfx.dest = 1, -1
    gfx.muladdrect(0,0,gfx.w, gfx.h,0,0,0,0) 
    local iw, ih = gfx.getimgdim(debugDrawZ)
    gfx.blit(debugDrawZ,1,0, 0, 0, gfx.w, gfx.h, 0, 0, gfx.w, gfx.h) 
    text('BUFFER '..debugDrawZ..' w:'..iw..' h:'..ih,0,0,200,20,0,{255,255,255},3)
  end
  
  if c >= 0 then reaper.runloop(runloop) end
  
end


activeMouseElement = nil
gfxWold, gfxHold = 0, 0
doArrange = true
runloop()


function Quit()
  d,x,y,w,h=gfx.dock(-1,0,0,0,0)
  reaper.SetExtState(sTitle,"wndw",w,true)
  reaper.SetExtState(sTitle,"wndh",h,true)
  reaper.SetExtState(sTitle,"dock",d,true)
  reaper.SetExtState(sTitle,"wndx",x,true)
  reaper.SetExtState(sTitle,"wndy",y,true)
  gfx.quit()
end
reaper.atexit(Quit)

end
